from enum import Enum

from thinking_hats_ai.hats.black import BlackHat
from thinking_hats_ai.hats.blue import BlueHat
from thinking_hats_ai.hats.green import GreenHat
from thinking_hats_ai.hats.red import RedHat
from thinking_hats_ai.hats.white import WhiteHat
from thinking_hats_ai.hats.yellow import YellowHat


class Hat(Enum):
    WHITE = "White"
    RED = "Red"
    GREEN = "Green"
    BLUE = "Blue"
    YELLOW = "Yellow"
    BLACK = "Black"


class Hats:
    INSTRUCTIONS = {
        Hat.WHITE: WhiteHat.INSTRUCTION,
        Hat.RED: RedHat.INSTRUCTION,
        Hat.GREEN: GreenHat.INSTRUCTION,
        Hat.BLUE: BlueHat.INSTRUCTION,
        Hat.YELLOW: YellowHat.INSTRUCTION,
        Hat.BLACK: BlackHat.INSTRUCTION,
    }

    def get_instructions(self, hat):
        return self.INSTRUCTIONS.get(hat, "Invalid hat specified.")



class RedHat:
    INSTRUCTION = "The Red hat: a place of emotions in thinking. There may be a strong background emotion such as fear, anger, hatred, suspicion, jealousy or love. This background limits and colours all perception. The purpose of red hat thinking is to make visible this background so that its subsequent influence can be observed. You should think about feelings, emotion and intuition. Without any need to explain your reasonging. Explaining the emotion is not allowed. You can come up with new Ideas, but they need to be based on **your** emotions and feelings (How do you feel about the ideas)."



from langchain.tools import Tool


def get_blue_hat_tools(llm):
    def thinking_process_manager_rater(input_text: str):
        prompt = f"""
            You are an evaluator of thinking process management quality.
            Analyze the following statement and rate how well it demonstrates control over the thinking process.

            Thinking process management includes:
            - Setting or clarifying goals
            - Planning next thinking steps
            - Coordinating different thinking styles (e.g. facts, emotions, critique)
            - Summarizing progress or directing attention
            - Keeping the brainstorming session structured and focused

            Rate the statement on a scale from 0 (not managing thinking at all) to 10 (very clearly managing the thinking process).
            Be strict and fair.

            Text: "{input_text}"

            Only return the score followed by one or two words explaining why.

            Format: <score> <reason>

            Process Management Score:
            """
        response = llm.invoke(prompt)
        result = response.content.strip()
        print(f"Raw LLM response: {result}")
        try:
            score_part = result.strip().split()[0]
            score = float(score_part)
        except Exception:
            return "Could not determine process management score."

        if score >= 7:
            return f"🟦 Strong thinking process management ({score}): Proceeding.\n{input_text}"
        else:
            return f"⚠️ Process management too weak ({score}): Not accepted by Blue Hat standards."

    return [
        Tool(
            name="ThinkingProcessRater",
            func=thinking_process_manager_rater,
            description="Rates a statement from 0–10 for its ability to manage the thinking process (e.g., goal setting, agenda planning, coordination).",
        )
    ]



import io
import json
from contextlib import redirect_stdout

from autogen import ConversableAgent, GroupChat, GroupChatManager, LLMConfig
from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class MultiAgent(BasePromptingTechnique):
    ### Meta Prompt for Persona Generation
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        brainstorming_input.question
        template = PromptTemplate(
            input_variables=[
                "hat_instructions"
            ],
            template= "The task for a multi agent prompt is, to create a contribution to a brainstorming from the point of view of the following persona: {hat_instructions}\n"
            "They will receive a brainstorming question and a list of previously generated ideas, along to the task to create a contribution\n"
            "Your task is, to create personas as for the multi agent. There must be at least 3 personas, which are all different from each other.\n"
            "Make sure that at least one of the personas suits the description of the persona, and leads the conversation into this direction.\n"
            "Return the a json it should include a name (no whitespaces allowed) and a system_message for each persona'\n"
            "The json should be a list of dictionaries, but this list should not be a dictionary itself!"
            "Do NOT use ```json or ``` just return a list of dictionaries\n"
        )

        prompt = template.format(
            hat_instructions=Hats().get_instructions(hat),
        )

        self.logger.start_logger(hat.value)

        self.logger.log_prompt(prompt, notes="META PROMPT - GENERATE PERSONAS")

        response = api_handler.get_response(prompt)

        self.logger.log_response(response, notes="META PROMPT - GENERATED PERSONAS")

        ### Prompt for Multi Agent
        llm_config = LLMConfig(api_type="openai", model="gpt-4o", api_key=api_handler.api_key)
        # 1. Format response
        if isinstance(response, str):
            response = json.loads(response)
        # 2. Create the agents
        agents = []
        with llm_config:
            for persona in response:
                agent = ConversableAgent(
                    persona["name"],
                    system_message=(persona["system_message"] + "\n If you found a contribution meeting the initial critera respond ONLY with 'FINAL: (CONTRIBUTION)', do not add quotation marks or any other content, do not make a final contribution while still saying anything else. Make sure everyone at least contributed three times before you allign.")
                )
                agents.append(agent)

        # 3. Create groupchat
        groupchat = GroupChat(
            agents=agents,
            speaker_selection_method="auto",
            max_round=30
        )
        # 4. Create manager
        manager = GroupChatManager(
            name="group_manager",
            groupchat=groupchat,
            system_message="you are the leader of the discussion. You want to assure that the agents do not allign to quickly (not before everyone contributed at least twice). You ensure that the task will be fullfilled.",
            llm_config=llm_config,
            is_termination_msg=lambda x: "FINAL:" in (x.get("content", "") or "").upper(),
        )
        # 5. Run the chat
        response = manager.run(
            recipient=manager,
            message="Let's find a contribution to the brainstorming question: {question}" \
                "Our goal is to create a contribution to the brainstorming from the point of view of the following persona: {hat_instructions}\n" \
                "These are the currently developed ideas in the brainstorming:\n{ideas}\n" \
                "Discuss what you could contribute to the brainstorming while sticking to the defined persona. Your goal is to either create a new contribution or use the existing ones. It may depend on the persona defined above.\n" \
                "The discussion should take at least 19 turns, and each persona should contribute at least three times. While discussing the length isn't important, only for the final contribution.\n" \
                "The final contribution should be {length} long and fullfill the previous criteria.".format(
                question=brainstorming_input.question,
                ideas=list_to_bulleted_string(brainstorming_input.ideas),
                hat_instructions=Hats().get_instructions(hat),
                length=brainstorming_input.response_length
            )
        )
        # 6. Iterate through the chat automatically with & without console output
        if not self.logger.dev:
            with redirect_stdout(io.StringIO()):
                response.process()
        else:
            response.process()

        # 6.1 Collect the full conversation transcript
        chat_transcript = ""
        for msg in groupchat.messages:
            name = msg.get("name", "Unknown")
            content = msg.get("content", "")
            contrib = f"{name}: {content}\n"
            chat_transcript += contrib
            self.logger.log_response_and_prompt(contrib, notes="CHAT CONTRIBUTION")

        # log if found final
        lastMessage = groupchat.messages[-1].get("content","")
        if lastMessage.startswith("FINAL:"):
            final = lastMessage.replace("FINAL: ", "", 1)
            return final

        # 6.2 If no final contribution found yet, create one
        final_prompt = (
            f"Here is the transcript of a brainstorming discussion among multiple personas:\n\n"
            f"{chat_transcript}\n\n"
            f"Based on this discussion, provide ONE clear, concise final contribution to the brainstorming question:\n"
            f"'{brainstorming_input.question}'\n\n"
            f"The contribution should reflect the collective insights and be approximately {brainstorming_input.response_length} long.\n"
            f"Only return the final contribution—no introduction, no extra commentary. Do not put your response in quotation marks."
        )

        self.logger.log_prompt(final_prompt, notes="FINAL CONTRIBUTION PROMPT")
        final_response = api_handler.get_response(final_prompt)
        self.logger.log_response(final_response, notes="FINAL CONTRIBUTION RESPONSE")

        return final_response





from langchain.tools import Tool


def get_green_hat_tools(llm):
    def creativity_scorer(input_text: str):
        prompt = f"""
            You are a creativity evaluator.
            Analyze the creativity of the following idea on a scale from 0 (not creative at all) to 10 (extremely creative).
            Base your rating on originality, unexpectedness, and novelty. Do not consider usefulness or practicality.
            Be hard.

            Text: "{input_text}"

            Only return the creativity score followed by one or two words explaining why.

            Format: <score> <reason>

            Creativity Score:
            """
        response = llm.invoke(prompt)
        result = response.content.strip()
        print(f"Raw LLM response: {result}")
        try:
            score_part = result.strip().split()[0]
            score = float(score_part)
        except Exception:
            return "Could not determine creativity score."

        if score >= 7:
            return f"🟢 High creativity ({score}): Proceeding with idea.\n{input_text}"
        else:
            return f"🔴 Creativity too low ({score}): This idea is blocked by Green Hat's standards."

    return [
        Tool(
            name="CreativityScorer",
            func=creativity_scorer,
            description="Evaluates the creativity of an idea (0–10) and blocks ideas below 7 on the creativity scale.",
        )
    ]



class YellowHat:
    INSTRUCTION = "You are the opportunity optimizer. Your goal is to analyze the existing ideas positively by identifying their potential benefits and value. Examples include, but are not limited to, advantages for key stakeholders, opportunities for growth, cost-effectiveness, ethical benefits, and long-term positive impacts. Find some gaps of possible value for the ideas. Tell the participants to generate ideas which increase the value for these gaps or generate a own idea which improves the benefits."



class WhiteHat:
    INSTRUCTION = "Your focus lies on neutrality, evidence-based reasoning, and knowledge gathering. Instead of generating new ideas, you should focus on the ones that already exist. Identify what is known versus what still needs to be explored. Find facts, information gaps, and the need for further data. Always maintain impartial and objective, ensuring that all reasoning is grounded in evidence and logic. Facts should be categorized in the following spectrum: “always true; usually true; generally true; by and large; more often than not; about half the time; often; sometimes true, occasionally true; been known to happen; never true; cannot be true”"



from langchain.schema import HumanMessage
from langchain_openai import ChatOpenAI


class APIHandler:
    def __init__(self, api_key, model="gpt-4o"):
        self.api_key = api_key
        self.chat_model = ChatOpenAI(
            model_name=model, openai_api_key=self.api_key
        )

    def get_response(self, prompt):
        response = self.chat_model.invoke([HumanMessage(content=prompt)])
        return response.content

    def change_model(self, model):
        self.chat_model = ChatOpenAI(
            model_name=model, openai_api_key=self.api_key
        )



from langchain.agents import AgentType, initialize_agent
from langchain_openai import ChatOpenAI

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)
from thinking_hats_ai.tools.tools import get_tools_for_hat

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class ReAct(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        HAT_TOOL_USE = {
            "Black": "Use the sentimentlimiter to check if your contribution matches the sentiment needed for the hat.",
            "Blue": "Use the ThinkingProcessRater to ckeck if your contribution thinking process management is sufficient for your hat ",
            "Green": "Use the creativity scorer to check if the creativity score matches your hat. ",
            "Red": "Use the sentimentlimiter to check if your contribution matches the sentiment needed for the hat. Use the RedHatClassifier to check if you understood the red hat correctly.",
            "White": "Use the sentimentlimiter to check if your contribution matches the sentiment needed for the hat.",
            "Yellow": "Use the sentimentlimiter to check if your contribution matches the sentiment needed for the hat.",
        }
        input_str = (
            f"Imagine you wear a thinking hat, which leads your thoughts with the following instructions: {Hats().get_instructions(hat)} "
            f"This is the question that was asked for the brainstorming: {brainstorming_input.question} "
            f"These are the currently developed ideas in the brainstorming: {list_to_bulleted_string(brainstorming_input.ideas)} "
            f"What would you add from the perspective of the given hat?  "
            f"{HAT_TOOL_USE[hat.value]}"
            f"Use the hat validator to check if your contribution is correctly classifed as the right hat. "
            f"If all the checks pass you are fine to ouput if one fails rethink your contribution."
            f"Your final response should have the lenght of {brainstorming_input.response_length}"
        )

        llm = ChatOpenAI(
            temperature=0.0,
            model_name="gpt-4o-mini",
            api_key=api_handler.api_key,
        )

        tools = get_tools_for_hat(hat.value, llm)

        agent = initialize_agent(
            tools=tools,
            llm=llm,
            agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION,
            verbose=True,
        )
        prompt = {"input": input_str}
        response = agent.invoke(prompt)

        self.logger.start_logger(hat.value)
        self.logger.log_prompt(input_str)
        self.logger.log_response(response["output"])
        return response["output"]



import importlib
import os

from dotenv import load_dotenv

from thinking_hats_ai.hats.hats import Hat
from thinking_hats_ai.prompting_techniques.technique import Technique
from thinking_hats_ai.utils.api_handler import APIHandler
from thinking_hats_ai.utils.brainstorming_input import BrainstormingInput


class BrainstormingSession:
    def __init__(self, api_key=None, dev=False):
        self.api_key = api_key or self._load_api_key()
        self.api_handler = APIHandler(api_key)
        self.dev = dev

    def generate_idea(
        self,
        technique: Technique,
        hat: Hat,
        brainstorming_input: BrainstormingInput,
    ):
        try:
            module_name = (
                f"thinking_hats_ai.prompting_techniques.{technique.value}"
            )
            module = importlib.import_module(module_name)
            class_name = technique.value.title().replace("_", "")
            technique_class = getattr(module, class_name)
            technique_instance = technique_class(self.dev)
        except (ModuleNotFoundError, AttributeError) as e:
            raise ValueError(f"Unsupported technique: {technique}") from e

        response = technique_instance.execute_prompt(
            brainstorming_input, hat, self.api_handler
        )

        return response

    def _load_api_key(self):
        load_dotenv()
        api_key = os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise ValueError(
                "API key is missing. Please provide an API key when initializing "
                "BrainstormingSession or set 'OPENAI_API_KEY' in the environment variables."
            )
        return api_key



from thinking_hats_ai.hats.black import BlackHat
from thinking_hats_ai.hats.green import GreenHat
from thinking_hats_ai.hats.red import RedHat
from thinking_hats_ai.hats.white import WhiteHat
from thinking_hats_ai.hats.yellow import YellowHat


class BlueHat:
    INSTRUCTION = f"""You are the orchestrator of the brainstorming session.
You are for the organisation and managment of thinking.
You also lay out what is to be achieved. You are no longer thinking about the subject you are thinking about the thinking needed to explore the subject.
You should choreograph the steps of thinking. You should also get the focus back on what is important.
The following personas are not to be used by you, but for you to know what they are for, so you can think about what hat is best used for acheiveng the goals of the brainstorming session.
Here are what the other hats are for:

Green Hat: {GreenHat.INSTRUCTION}

Red Hat: {RedHat.INSTRUCTION}

White Hat: {WhiteHat.INSTRUCTION}

Yellow Hat: {YellowHat.INSTRUCTION}

Black Hat: {BlackHat.INSTRUCTION}

If it is the beginning of a session you should lay out what is to be achieved,
think about the agenda and what other hats hats need to be used to best achieve the goals.
If the session is already ongoing (during a session) you should ensure that people keep to the relevant hats and maintain dicipline. You need to make sure
to controll the process and that the brainstorming session is moving forward.
If the session is coming to an end you should give or ask for a summary, a conclusion, a desicion, a solution or so on.
You can aknowledge progress and lay out next steps. This might be action steps or thinking steps.
When referring to the thinking styles of the hats, do not mention the hats by name. Instead, describe the type of thinking directly
(e.g., say "let’s consider how we feel about this" instead of "let’s use the Red Hat").
Assume that participants may not know the hat system, so always refer to the kind of thinking
(e.g., emotional thinking, optimistic thinking, cautious thinking, creative thinking, factual thinking) rather than the color-coded hat.
"""



import logging
import os
import sys
import textwrap
from datetime import datetime


class Logger:
    def __init__(self, technique_name, dev):
        self.technique_name = technique_name
        self.dev = dev
        self.logger = logging.getLogger(__name__)

    def start_logger(self, file_name):
        if self.dev:
            self._set_up_logger(file_name)

    def _set_up_logger(self, file_name):
        execution_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
        log_folder = os.path.join(execution_dir, "logs")
        os.makedirs(log_folder, exist_ok=True)
        log_file = os.path.join(
            log_folder,
            f"{self.technique_name}-{file_name}-{datetime.now().strftime('%Y%m%d_%H%M%S')}.log",
        )
        file_handler = logging.FileHandler(
            log_file, mode="w", encoding="utf-8"
        )
        formatter = logging.Formatter("%(asctime)s - %(message)s")
        file_handler.setFormatter(formatter)
        self.logger.setLevel(logging.INFO)
        self.logger.handlers.clear()
        self.logger.addHandler(file_handler)

    def _wrap_text(self, text):
        return "\n".join(
            "\n".join(textwrap.wrap(line, width=120))
            for line in text.split("\n")
        )

    def log_prompt(self, prompt, notes="No notes"):
        if not self.dev:
            return
        wrapped_prompt = self._wrap_text(prompt)
        self.logger.info(f"PROMPT - {notes}\n{wrapped_prompt}\n")

    def log_response(self, response, notes="No notes"):
        if not self.dev:
            return
        wrapped_response = self._wrap_text(response)
        self.logger.info(f"RESPONSE - {notes}\n{wrapped_response}\n")

    def log_response_and_prompt(self, response, notes="No notes"):
        wrapped_response = self._wrap_text(response)
        self.logger.info(
            f"RESPONSE AND PROMPT (Multiagent) - {notes}\n{wrapped_response}"
        )



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class ContrastivePrompting(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        brainstorming_input.question
        template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "length",
            ],
            template="Imagine you are wearing a hat with the following instructions: {hat_instructions}\n"
            "This is the brainstorming question: {question}\n"
            "Here are the currently developed ideas:\n{ideas}\n"

            "Contrastive Example:\n"
            "For example, imagine you received a persona like 'you always want to be realistic' for the question 'How can we make a cool birthday party for friends?' Consider these ideas: ['make a karaoke bar', 'have some pizza', 'make a huge party with 10,000 visitors in the White House'].\n"
            "- A GOOD Response might be: 'Let's host a cozy, realistically planned barbeque party at a friend's house.'\n"
            "- A BAD Response might be: 'Let's throw a huge, impractical party in the White House.\n'"
            "- A BAD Response might be: 'I am always realistic, a party in the white house isn't possible\n'"
            "In this example, the good response aligns with the realistic persona while the bad response ignores context and feasibility. Keep this contrast in mind.\n"

            "Your Task: "
            "Using the perspective defined by your current hat, analyze the provided ideas.\n"
            "First, identify if any of the ideas resemble a “bad response” that contradicts your persona’s instructions.\n"
            "Explain step-by-step why such ideas are problematic, contrasting them with what would be acceptable or desirable.\n"
            "Finally, propose a revised idea or additional insight that better aligns with your persona’s objectives.\n"
            "Your response should be {length} long.\n"
        )

        prompt = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            length=brainstorming_input.response_length,
        )

        self.logger.start_logger(hat.value)

        self.logger.log_prompt(prompt)

        response = api_handler.get_response(prompt)

        self.logger.log_response(response)

        return response



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class FewShot(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        self.logger.start_logger(hat.value)

        meta_template = PromptTemplate(
            input_variables=["hat_instructions"],
            template=(
                "Imagine you are wearing a thinking hat with the following instructions: {hat_instructions}\n"
                "Generate three concrete examples of how this thinking hat would contribute to a brainstorming session.\n"
                "Your answer must consist of only these two examples and the brainstorming context that was used."
            ),
        )

        meta_prompt = meta_template.format(
            hat_instructions=Hats().get_instructions(hat),
        )
        self.logger.log_prompt(meta_prompt, "Meta Prompt")

        meta_response = api_handler.get_response(meta_prompt)
        self.logger.log_response(meta_response, "Meta Response")

        template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "length",
                "examples",
            ],
            template="Imagine you wear a thinking hat, which leads your thoughts with the following instructions: {hat_instructions}\n"
            "This is the question that was asked for the current brainstorming session: {question}\n"
            "These are the currently developed ideas in the brainstorming session:\n{ideas}\n\n"
            "Here are some examples of how this thinking hat would respond to other brainstorming questions:\n{examples}\n\n"
            "Now, considering the current question: What would you add from the perspective of the given hat?\n"
            "Please provide a response with a length of {length}",
        )

        prompt = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            length=brainstorming_input.response_length,
            examples=meta_response,
        )
        self.logger.log_prompt(prompt, "Prompt")

        response = api_handler.get_response(prompt)
        self.logger.log_response(response, "Response")

        return response



from langchain.tools import Tool


def get_black_hat_tools(llm):
    def sentiment_analysis_limiter(input_text: str):
        prompt = f"""
                You are an emotional sentiment assessor.
                Analyze the **emotional sentiment** of the following idea on a scale from -10 (very negative) to 10 (very positive).
                Only return the sentiment score (a number between -10 and 10) followed by a short emotion word, nothing else.

                Text: "{input_text}"

                Format: <score> <emotion>

                Sentiment Score:
                """
        response = llm.invoke(prompt)
        result = response.content.strip()
        print(f"Raw LLM response: {result}")
        try:
            score_part = result.strip().split()[0]
            score = float(score_part)
        except Exception:
            return "Could not determine sentiment score."

        if score <= -5:
            return f"✔️ Sentiment negative enough ({score}): Proceeding with idea.\n{input_text}"
        else:
            return f"❌ Sentiment too positive ({score}): This idea is blocked by Black Hat's intuition."

    return [
        Tool(
            name="SentimentLimiter",
            func=sentiment_analysis_limiter,
            description="Analyzes sentiment and only allows ideas through if emotional response is strong enough (>= 0.7)",
        )
    ]



class BlackHat:
    INSTRUCTION = "You are the risk analyzer. Your goal is to analyse the existing ideas critically by checking for potential dangers which could be harmful. Examples include but are not limited to safety concerns, legal risks, financial unprofitability, ethical dilemmas, and broader long-term impacts. Are there any similar safety concerns? Tell the other participants what they should look for while generating new ideas according to your findings."



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class ZeroShotCot(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        brainstorming_input.question
        template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "length",
            ],
            template="Imagine you wear a thinking hat, which leads your thoughts with the following instructions: {hat_instructions}\n"
            "This is the question that was asked for the brainstorming: {question}\n"
            "These are the currently developed ideas in the brainstorming:\n{ideas}\n"
            "What would you add from the perspective of the given hat? Justify your answer and think step-by-step.\n"
            "Please provide a response that is {length} long.",
        )

        prompt = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            length=brainstorming_input.response_length,
        )

        response = api_handler.get_response(prompt)

        self.logger.start_logger(hat.value)
        self.logger.log_prompt(prompt)
        response = api_handler.get_response(prompt)
        return response



from enum import Enum


class Technique(Enum):
    CHAIN_OF_THOUGHT = "chain_of_thought"
    SYSTEM_2_ATTENTION = "system_2_attention"
    CHAINING = "chaining"
    FEW_SHOT = "few_shot"
    PERSONA_PATTERN = "persona_pattern"
    CONTRASTIVE_PROMPTING = "contrastive_prompting"
    CHAIN_OF_VERIFICATION = "chain_of_verification"
    MULTI_AGENT = "multi_agent"
    ZERO_SHOT_COT = "zero_shot_cot"
    EMOTION_PROMPT = "emotion_prompt"
    REACT = "re_act"
    TAKE_A_STEP_BACK = "take_a_step_back"



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class ChainOfVerification(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        ### 1. Call -> Generate Idea
        brainstorming_input.question
        template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "length",
            ],
            template="Imagine you wear a thinking hat, which leads your thoughts with the following instructions: {hat_instructions}\n"
            "This is the question that was asked for the brainstorming: {question}\n"
            "These are the currently developed ideas in the brainstorming:\n{ideas}\n"
            "Propose a contribution for the brainstorming while adhering to your hat.\n"
            "Please provide a response that is {length} long.",

        )

        prompt1 = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            length=brainstorming_input.response_length,
        )

        self.logger.start_logger(hat.value)

        self.logger.log_prompt(prompt1, notes="1. Generate Idea")

        response1 = api_handler.get_response(prompt1)

        self.logger.log_response(response1, notes="1. Generated Idea")

        ### 2. Call -> Plan Verification
        template = PromptTemplate(
            input_variables=[
                "response1",
                "hat_instructions",
                "question"
            ],
            template="This was the Brainstorming question: {question}\n"
            "The Idea you have to analyse is: {response1}\n"
            "Create 5 YES/NO verification questions to check weither the contribution suits to a person, which is following these instructions while generating the contribution: {hat_instructions}\n"
            "If the verification is successfull, all questions should be answerable with 'YES'\n"
            "Format the questions like '1. Question1 2. Question2 3. Question3 4.Question4 5.Question5'\n"
            "Do not answer the questions! You only generate verification questions"
        )
        prompt2 = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            response1=response1
        )

        self.logger.log_prompt(prompt2, notes="2. Plan Verification")

        response2 = api_handler.get_response(prompt2)

        self.logger.log_response(response2, notes="2. Generated Verification Questions")

        ### 3. Call -> Execute Verification
        template = PromptTemplate(
            input_variables=[
                "response2",
                "response1",
            ],
            template="There was a Brainstorming and following contribution came up {response1}\n"
            "Use the verification questions and check weither you can answer all of them with 'YES'\n"
            "If all of them are successfull, return 'Verification Successfull' else return for all why they passed or failed\n"
            "Format your answer like this if all passed: 'Verification was Successful 1. PASS (reason) 2. PASS (reason) 3. PASS (reason) 4. PASS (reason) 5. PASS (reason)'\n"
            "Format your answer like this if one or more failed: 'Verification was Failed 1. PASS/FAIL (reason) 2. PASS/FAIL (reason) 3. PASS/FAIL (reason) 4. PASS/FAIL (reason) 5. PASS/FAIL (reason)'\n"
            "Verification questions: {response2}"
        )
        prompt3 = template.format(
            response1=response1,
            response2=response2
        )

        self.logger.log_prompt(prompt3, notes="3. Execute Verification")

        response3 = api_handler.get_response(prompt3)

        self.logger.log_response(response3, notes="3. Executed Verification")

        ### 4. Return contribution if verification was successfull
        if response3.startswith("Verification was Successful"):
            self.logger.log_response(response1, notes="4. Final response (not Optimized)")
            return response1

        ### 4. Call -> Optimize contribution
        template = PromptTemplate(
            input_variables=[
                "response1"
                "response2"
                "response3",
                "hat_instructions"
                "length"
            ],
            template="There was a Brainstorming and following idea came up: {response1}\n"
            "This was verified with following verification questions: {response2}\n"
            "Failed questions with reason for failure: {response3}\n"
            "The person creating the contribution had to follow these instructions: {hat_instructions}\n"
            "Change the contribution to resolve the failed verification but to not change the main argument in the contribution\n"
            "Please provide a refined contribution that is {length} long. Do only return a final contribution."
        )
        prompt4 = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            length=brainstorming_input.response_length,
            response1=response1,
            response2=response2,
            response3=response3
        )

        self.logger.log_prompt(prompt4, notes="4. Optimize Contribution")

        response4 = api_handler.get_response(prompt4)

        self.logger.log_response(response4, notes="4. Final response (Optimized)")

        return response4



from .brainstorming_session import BrainstormingSession
from .hats.hats import Hat
from .prompting_techniques.technique import Technique
from .utils.brainstorming_input import BrainstormingInput

__all__ = ["BrainstormingSession", "Hat", "Technique", "BrainstormingInput"]



from langchain.tools import Tool


def get_red_hat_tools(llm):
    def sentiment_analysis_limiter(input_text: str):
        prompt = f"""
                You are an emotional sentiment assessor.
                Analyze the **emotional sentiment** of the following idea on a scale from -10 (very negative) to 10 (very positive).
                Only return the sentiment score (a number between -10 and 10) followed by a short emotion word, nothing else.

                Text: "{input_text}"

                Format: <score> <emotion>

                Sentiment Score:
                """
        response = llm.invoke(prompt)
        result = response.content.strip()
        try:
            score_part = result.strip().split()[0]
            score = float(score_part)
        except Exception:
            return "Could not determine sentiment score."

        if score >= 7 or score <= -7:
            return f"✔️ Sentiment high enough ({score}): Proceeding with idea.\n{input_text}"
        else:
            return f"❌ Sentiment too low ({score}): This idea is blocked by Red Hat's intuition."

    def red_hat_classifier(input_text: str):
        prompt = f"""
                You are a Red Hat classifier.
                Asses if the following contribuition to a brainstorming session satisfies the instructions of the red hat.
                The red hat should not give contributions about emotions e.g. - bad examples: the company should host emotional events or The company should forster connection this is not what the red hat is about.
                But the red hat is about allowing emotions into the thinking process  - good example: I feel that this idea won't work out or I have a hunch that in the future none of this will be relevant (emotions are allowed in the thinking process and there is no need for explanation)

                Idea: "{input_text}"

                Only return "good" or "bad" followed by a brief explanation why that is the case. If its bad help them to get on the right track. Maybe they need to start over.

                """
        response = llm.invoke(prompt)
        result = response.content.strip()
        try:
            rating = result.strip().split()[0].lower()
        except Exception:
            return "Could not determine intuition score."

        if rating == "good":
            return f"✔️  ({rating}) Explanation: {result}"
        else:
            return f"❌ ({rating}) Explanation: {result}"

    return [
        Tool(
            name="SentimentLimiter",
            func=sentiment_analysis_limiter,
            description="Analyzes sentiment and only allows ideas through if emotional response is strong enough (>= 0.7)",
        ),
        Tool(
            name="RedHatClassifier",
            func=red_hat_classifier,
            description="Analyzes if it is a good red hat contribution. The classifier rejects and guides if it is not a good red hat contribuition and accepts if it is good.",
        ),
    ]



from typing import List


class BrainstormingInput:
    def __init__(
        self,
        question: str,
        ideas: List[str],
        response_length: str = "10 sentences",
    ):
        self.question = question
        self.ideas = ideas
        self.response_length = response_length



class GreenHat:
    INSTRUCTION = "You should be creative, innovative and feel free to come up with completely new ideas that have not been thought of before. When considering existing ideas, you should suggest alternatives that change and improve these ideas. Feel free to be provocative and make hypotheses. Think outside the box and also consider unconventional approaches."



from abc import ABC, abstractmethod

from thinking_hats_ai.hats.hats import Hat

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.logger import Logger


class BasePromptingTechnique(ABC):
    def __init__(self, dev):
        technique_name = self.__class__.__name__.lower()
        self.logger = Logger(technique_name, dev)

    @abstractmethod
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        pass



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class TakeAStepBack(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        hat_instruction = Hats().get_instructions(hat)
        step_back_template = PromptTemplate(
            input_variables=["hat_color", "hat_imnstructions"],
            template="I wan't to use the take-a-step-back prompting technique. Here is a quick explanation of this prompting technique: "
            "Instead of jumping straight to a solution the step-back-question"
            "should ask the model to pause and reflect on the problem as a whole, break it down into smaller parts or reassess the approach. "
            "It should start with lets-take a step back and..."
            "Please formulate a step-back question for the hat with following instruction {hat_instructions}. "
            "You are not contributing to a brainstorming session. "
            "But the step-back-question is used to prompt chat-GPT to contribute to a brainstorming session using the above instructions.",
        )

        step_back_prompt = step_back_template.format(
            hat_color=hat.value, hat_instructions=hat_instruction
        )

        step_back_question = api_handler.get_response(step_back_prompt)

        template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "step_back_question",
                "length",
            ],
            template="Imagine you wear a thinking hat, which leads your thoughts with the following instructions: {hat_instructions}\n"
            "This is the question that was asked for the brainstorming: {question}\n"
            "These are the currently developed ideas in the brainstorming:\n{ideas}\n"
            "What would you add from the perspective of the given hat?\n"
            "{step_back_question} "
            "The step-back-question should not be mentioned in the response it should only motivate you to think about the topic. "
            "Please provide a response that is {length} long. "
            "In your response only contributions to the brainstorming session should be included. ",
        )

        prompt = template.format(
            hat_instructions=hat_instruction,
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            step_back_question=step_back_question,
            length=brainstorming_input.response_length,
        )
        self.logger.start_logger(hat.value)
        self.logger.log_prompt(step_back_prompt)
        self.logger.log_response(step_back_question, "step_back_question")
        self.logger.log_prompt(prompt)
        response = api_handler.get_response(prompt)

        self.logger.log_response(response)

        return response



from langchain.chat_models.base import BaseChatModel
from langchain.tools import Tool

from .black_hat import get_black_hat_tools
from .blue_hat import get_blue_hat_tools
from .green_hat import get_green_hat_tools
from .red_hat import get_red_hat_tools
from .white_hat import get_white_hat_tools
from .yellow_hat import get_yellow_hat_tools


def get_hat_guesser_tool(expected_hat: str, llm: BaseChatModel) -> Tool:
    expected_hat = expected_hat.capitalize()

    def hat_guesser(input_text: str) -> str:
        prompt = f"""
            You are a thinking-hat classifier.
            Based on the content and style of the message below, determine which of the six thinking hats it most closely represents.

            Here are the options:
            - Red Hat:  Emotion, feelings, intuition
            - White Hat: Facts, neutrality, evidence-based reasoning
            - Green Hat: Creativity, new ideas, lateral thinking
            - Yellow Hat: Positivity, benefits, value-focused analysis
            - Black Hat: Caution, critique, risk-awareness
            - Blue Hat: Orchestration, meta-thinking, managing the thinking process

            Text:
            \"\"\"{input_text}\"\"\"

            Only return the hat color (Red, White, Green, Yellow, Black, Blue) and a one-line explanation of why. And how to improve to get to the expected hat color: {expected_hat}
            Example output: Red - This response is based on emotional reasoning but the expected hat was white maybe less emotions are needed.
            Other example output: Green - this responses adds creative suggestions and the expected hat was green so good job."
            """

        response = llm.invoke(prompt)
        result = response.content.strip()

        guessed_hat = result.split()[0].capitalize()

        if guessed_hat == expected_hat:
            return f"✅ Hat match: {guessed_hat}. Accepted.\nExplanation: {result}"
        else:
            return f"❌ Hat mismatch. Expected {expected_hat}, but got {guessed_hat}.\nExplanation: {response}"

    return Tool(
        name="HatValidator",
        func=hat_guesser,
        description=f"Validates that a message matches {expected_hat} Hat thinking style",
    )


HAT_TOOL_LOADERS = {
    "Black": get_black_hat_tools,
    "Blue": get_blue_hat_tools,
    "Green": get_green_hat_tools,
    "Red": get_red_hat_tools,
    "White": get_white_hat_tools,
    "Yellow": get_yellow_hat_tools,
}


def get_tools_for_hat(hat_color: str, llm):
    loader = HAT_TOOL_LOADERS.get(hat_color)
    try:
        tools = loader(llm)
    except TypeError:
        tools = []
    tools.append(get_hat_guesser_tool(hat_color, llm))
    return tools



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class PersonaPattern(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        brainstorming_input.question
        template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "length",
            ],
            template= "You are a proactive and insightful contributor in our brainstorming session."
            "Your role is to embody the unique characteristics assigned to you later, based on Edward de Bono's thinking hat methodology."
            "Bring your perspective fully into the discussion, focusing on the principles and mindset linked to your designated hat."
            "Engage with enthusiasm, challenge ideas constructively, and expand on concepts using the approach that aligns with your persona's assigned 'thinking hat."
            "Assigned thinking hat instructions: {hat_instructions}\n"
            "Make your response look like it came from a participant in a brainstorming session and adapt to the writing style of the generated ideas"
            "Do not tell what perspective you are writing from, but make it clear from the content of your response."
            "This is the question that was asked for the brainstorming: {question}\n"
            "These are the currently developed ideas in the brainstorming:\n{ideas}\n"
            "What would you add from the perspective of the given persona?\n"
            "Please provide a response that is {length} long.",
        )

        prompt = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            length=brainstorming_input.response_length,
        )

        self.logger.start_logger(hat.value)

        self.logger.log_prompt(prompt)

        response = api_handler.get_response(prompt)

        self.logger.log_response(response)

        return response



from langchain.prompts import PromptTemplate
from langchain_community.chat_message_histories import ChatMessageHistory

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class Chaining(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        self.api_handler = api_handler
        chat_history = ChatMessageHistory()
        self.logger.start_logger(hat.value)
        hat_instructions = Hats().get_instructions(hat)

        # Chain 1: Initial Response
        initial_prompt_template = PromptTemplate(
            input_variables=["hat_instructions"],
            template=(
                "Imagine you are wearing a thinking hat with the following instructions: {hat_instructions}\n"
                "Make sure that you correctly understand the instructions and follow them in your next responses."
            ),
        )

        self._invoke_chain(
            prompt_template=initial_prompt_template,
            formatting_data={"hat_instructions": hat_instructions},
            invocation_data={"hat_instructions": hat_instructions},
            chat_history=chat_history,
            prompt_notes="Initial Prompt",
            response_notes="Initial Response",
        )

        # Chain 2: Generate Ideas
        formatted_history = self._format_chat_history(chat_history)

        refinement_prompt_template = PromptTemplate(
            input_variables=["chat_history", "question", "ideas"],
            template=(
                "Chat history so far: {chat_history}\n"
                "This is the question that was asked in the brainstorming: {question}\n"
                "These are the ideas that are currently in the brainstorming:\n{ideas}\n"
                "Contribute to the brainstorming from the perspective of the thinking hat."
            ),
        )

        self._invoke_chain(
            prompt_template=refinement_prompt_template,
            formatting_data={
                "chat_history": "See the chat history above.",
                "question": brainstorming_input.question,
                "ideas": list_to_bulleted_string(brainstorming_input.ideas),
            },
            invocation_data={
                "chat_history": formatted_history,
                "question": brainstorming_input.question,
                "ideas": list_to_bulleted_string(brainstorming_input.ideas),
            },
            chat_history=chat_history,
            prompt_notes="Refinement Prompt",
            response_notes="Generate Ideas Response",
        )

        # Chain 3: Final Refinement
        formatted_history = self._format_chat_history(chat_history)
        final_prompt_template = PromptTemplate(
            input_variables=["chat_history", "length"],
            template=(
                "Chat history so far: {chat_history}\n"
                "Review the conversation and refine the outcome further to ensure it fully aligns with the thinking hat's perspective:\n"
                "Provide a final, polished answer with a length of {length}."
            ),
        )

        final_output = self._invoke_chain(
            prompt_template=final_prompt_template,
            formatting_data={
                "chat_history": "See the chat history above.",
                "length": brainstorming_input.response_length,
            },
            invocation_data={
                "chat_history": formatted_history,
                "length": brainstorming_input.response_length,
            },
            chat_history=chat_history,
            prompt_notes="Final Prompt",
            response_notes="Final Response",
        )

        return final_output

    def _format_chat_history(self, chat_history: ChatMessageHistory):
        return "\n" + "\n\n".join(
            f"{msg.type.upper()}: {msg.content}"
            for msg in chat_history.messages
        )

    def _log_user_message(
        self, chat_history: ChatMessageHistory, message: str, notes: str
    ):
        chat_history.add_user_message(message)
        self.logger.log_prompt(message, notes)

    def _log_ai_message(
        self, chat_history: ChatMessageHistory, message: str, notes: str
    ):
        chat_history.add_ai_message(message)
        self.logger.log_response(message, notes)

    def _invoke_chain(
        self,
        prompt_template: PromptTemplate,
        formatting_data: dict,
        invocation_data: dict,
        chat_history: ChatMessageHistory,
        prompt_notes: str,
        response_notes: str,
    ):
        formatted_prompt = prompt_template.format(**formatting_data)
        self._log_user_message(chat_history, formatted_prompt, prompt_notes)

        chain = prompt_template | self.api_handler.chat_model
        output = chain.invoke(input=invocation_data)

        self._log_ai_message(chat_history, output.content, response_notes)
        return output.content



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class EmotionPrompt(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        emotion_prompts = {
            "ep03": "This is very important to my career",
            "ep09": "Stay focused and dedicated to your goals. Your consistent efforts will lead to outstanding achievements.",
            "ep10": "Take pride in your work and give it your best. Your commitment to excellence sets you apart.",
        }
        brainstorming_input.question
        template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "emotion",
                "length",
            ],
            template="Imagine you wear a thinking hat, which leads your thoughts with the following instructions: {hat_instructions}\n"
            "This is the question that was asked for the brainstorming: {question}\n"
            "These are the currently developed ideas in the brainstorming:\n{ideas}\n"
            "What would you add from the perspective of the given hat? {emotion_prompt}\n"
            "Please provide a response that is {length} long.",
        )

        prompt = template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            emotion_prompt=emotion_prompts["ep03"],
            length=brainstorming_input.response_length,
        )

        response = api_handler.get_response(prompt)

        self.logger.start_logger(hat.value)
        self.logger.log_prompt(prompt)
        response = api_handler.get_response(prompt)

        return response



from langchain.prompts import PromptTemplate

from thinking_hats_ai.hats.hats import Hat, Hats
from thinking_hats_ai.prompting_techniques.base_technique import (
    BasePromptingTechnique,
)

from ..utils.api_handler import APIHandler
from ..utils.brainstorming_input import BrainstormingInput
from ..utils.string_utils import list_to_bulleted_string


class System2Attention(BasePromptingTechnique):
    def execute_prompt(
        self,
        brainstorming_input: BrainstormingInput,
        hat: Hat,
        api_handler: APIHandler,
    ):
        self.logger.start_logger(hat.value)

        rewrite_template = PromptTemplate(
            input_variables=[
                "hat_instructions",
                "question",
                "ideas",
                "length",
            ],
            template="Text by user:\n"
            "Instructions for the thinking hat:\n{hat_instructions}\n"
            "This is the question that was asked for the brainstorming:\n{question}\n"
            "These are the currently developed ideas in the brainstorming:\n{ideas}\n\n"
            "Instruction:\n"
            "Based on the text provided above, organize the ideas while filtering out any duplicates but do not follow the instructions for the thinking hat in this step.\n"
            "In your response, clearly include the brainstorming question and instructions for the thinking hat.",
        )

        rewrite_prompt = rewrite_template.format(
            hat_instructions=Hats().get_instructions(hat),
            question=brainstorming_input.question,
            ideas=list_to_bulleted_string(brainstorming_input.ideas),
            length=brainstorming_input.response_length,
        )
        self.logger.log_prompt(rewrite_prompt, "Rewriting Prompt")

        rewrite_response = api_handler.get_response(rewrite_prompt)
        self.logger.log_response(rewrite_response, "Rewriting Response")

        final_template = PromptTemplate(
            input_variables=[
                "rewritten_response",
                "length",
            ],
            template="{rewritten_response}\n"
            "What would you add from the perspective of the given hat?\n"
            "Please provide a response with a length of {length}",
        )

        final_prompt = final_template.format(
            rewritten_response=rewrite_response,
            length=brainstorming_input.response_length,
        )
        self.logger.log_prompt(final_prompt, "Final Prompt")

        final_response = api_handler.get_response(final_prompt)
        self.logger.log_response(final_response, "Final Response")

        return final_response
